/*
 * This version of the simulator models semaphores in which the
 * following changes have been made:
 *
 * A buzz lock protects all semaphore operations from execution
 * of other CPUs.  The lock is released at only a few specified
 * points.
 *
 * Interrupt lockout protects down() operations from interrupt
 * code.
 */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

/*
 * Set to 1 for normalize function to be atomic, set to zero for
 * non-atomic.  Non-atomic produces errors.
 */
#ifndef ATOMIC_NORMALIZE
#define	ATOMIC_NORMALIZE	1
#endif

#define	N_CPUS		8
int	n_cpus ;

/*
 * The simulated semaphore has count and waiting field just like
 * the real kernel semaphores.  It also has a state vector each slot
 * of which is used by a different simulated CPU to track the stages
 * of semaphore processing.  An interrupt is viewed as a separate CPU
 * that only does V operations.
 *
 * For simulation purposes, there is only one semaphore and multiple
 * users of it.
 */
typedef struct semaphore
{
	int	count ;			/* counter */
	int	waiting ;		/* counter */
	int	lock ;
	int	state[N_CPUS] ;		/* state of operation */
	int	number[N_CPUS] ;	/* saved 'waiting' after exchg */

} semaphore_t ;

/*
 * States of P operation
 */
#define	IDLE		0		/* no P/V() operation in progress */
#define	P_START		1		/* start a P() operation */
#define	P_GET_LOCK1	2		/* P: get buzz lock*/
#define	P_PROCESS	3		/* P: do most of processing */
#define	P_TEST1		4		/* P: test count+waiting */
#define	P_WAITING	5		/* P: give lock and wait on semaphore */
#define	P_TEST2		6		/* P: test to see if keep waiting */
#define	P_WAIT_DONE	7		/* P: wait complete */
#define	P_GET_LOCK2	8		/* P: get lock again */
#define	P_ACQUIRED	9		/* P: semaphore acquired */
#define	P_RETRY		10		/* P: retry semaphore */
/*
 * States of V operation
 */
#define	V_START		20		/* start a V() operation */
#define	V_INCR_CNT	21		/* V: incr count field */
#define	V_GET_LOCK	22		/* V: get buzz lock*/
#define	V_PROCESS	23		/* V: do most of processing */
#define	V_WAKEUP	24		/* V: wake up all waiters */
#define	V_COMPLETE	25		/* V: operation complete */

#define	N_STATES	26		/* number of states */

semaphore_t	sem ;			/* THE semaphore */


typedef struct
{
	int	cpu ;
	int	old_state ;
	int	new_state ;
	int	count ;
	int	waiting ;
	int	number ;

} state_hist_t ;


#define	N_STATE_HIST	1024
state_hist_t	state_history[N_STATE_HIST] ;
int		state_hist_inx ;
state_hist_t	cpu_history[N_CPUS] ;
int		state_map[N_STATES] ;

int		iterations ;
int		correct ;
int		correct_cnt ;
int		incorrect_cnt ;
int		cycle_cnt ;
int		n_examples = 2 ;	/* number of examples to print */

char		cpu_path[2000] = " " ;

char	*state_name(int state)
{
	static char	ascii[32] ;

	switch (state)
	{
	case IDLE:		return("IDLE");
	case P_START:		return("P_START");
	case P_GET_LOCK1:	return("P_GET_LOCK1") ;
	case P_PROCESS:		return("P_PROCESS") ;
	case P_TEST1:		return("P_TEST1");
	case P_WAITING:		return("P_WAITING");
	case P_TEST2:		return("P_TEST2");
	case P_WAIT_DONE:	return("P_WAIT_DONE");
	case P_GET_LOCK2:	return("P_GET_LOCK2") ;
	case P_ACQUIRED:	return("P_ACQUIRED");
	case P_RETRY:		return("P_RETRY");
	case V_START:		return("V_START");
	case V_INCR_CNT:	return("V_INCR_CNT");
	case V_GET_LOCK:	return("V_GET_LOCK");
	case V_PROCESS:		return("V_PROCESS");
	case V_WAKEUP:		return("V_WAKEUP");
	case V_COMPLETE:	return("V_COMPLETE");

	default:
		sprintf(ascii, "State=%d", state) ;
		return(ascii) ;
	}
}

int get_lock(void)
{
    if (sem.lock == 0)
    {
	sem.lock = 1 ;
	return(1) ;			/* got lock */
    }

    return(0) ;				/* did not get lock */
}

void give_lock(void)
{
    sem.lock = 0 ;
}

int decr_cnt(void)
{
	return(--sem.count) ;
}

int incr_wtg(void)
{
	return(++sem.waiting) ;
}

int test1(void)
{
	return(sem.count + sem.waiting <= 0) ;
}

int test2(void)
{
	return(sem.count < 0) ;
}

int norm_exchg(void)
{
	int	tmp ;

	tmp = sem.waiting ;
	sem.waiting = 0 ;
	return(tmp) ;
}

int norm_add(int number)
{
	sem.count += number ;
}

int incr_cnt(void)
{
	return(++sem.count) ;
}

void save_state_hist(state_hist_t *p, int cpu,
		     int old_state, int new_state, int *inx)
{
	p->cpu       = cpu ;
	p->old_state = old_state ;
	p->new_state = new_state ;
	p->count     = sem.count ;
	p->waiting   = sem.waiting ;
	p->number    = sem.number[cpu] ;

	if (inx != NULL)
		*inx = (*inx + 1) % N_STATE_HIST ;

}

void set_state(int cpu, int state)
{
	save_state_hist(&state_history[state_hist_inx], 
			cpu, sem.state[cpu], state, &state_hist_inx) ;

	save_state_hist(&cpu_history[cpu], 
			cpu, sem.state[cpu], state, NULL) ;

	state_map[sem.state[cpu]]-- ;		/* one fewer in that state */
	state_map[state]++ ;			/* one more in this state */
	sem.state[cpu] = state ;
}

void print_state_hist(state_hist_t *p, int index)
{
	if (p->old_state == -1) return ;

	printf("cpu %d: index=%04d count=%d waiting=%d number=%d %s",
		p->cpu,
		++index,
		p->count, p->waiting, p->number,
		state_name(p->old_state)) ;

	if (p->new_state >= 0)
		printf(" -> %s", state_name(p->new_state)) ;

	printf("\n") ;
} 

void dump_state_hist(void)
{
#if 0
	int	i ;
	int	n ;

	for (n = N_STATE_HIST, i = state_hist_inx;
	     n > 0;
	     n--, i = (i + 1) % N_STATE_HIST)
        {
		print_state_hist(&state_history[i]) ;
        }
#else
	int	i ;

	for (i = 0;
	     i < state_hist_inx;
	     i++
	    )
        {
		print_state_hist(&state_history[i], i) ;
        }
#endif

	fflush(stdout) ;
}

void dump_cpu_hist(void)
{
	int	i ;

	for (i = 0; i < n_cpus; i++)
        {
		print_state_hist(&cpu_history[i], i) ;
        }

	fflush(stdout) ;
}

void init_state_hist(void)
{
	int	i ;

	for (i = 0; i < N_STATE_HIST; i++)
	{
		state_history[i].old_state = -1 ;
		state_history[i].new_state = -1 ;
	}

	state_hist_inx = 0 ;
	memset(state_map, 0, sizeof(state_map)) ;
}

void init_cpu_hist(void)
{
	int	i ;

	for (i = 0; i < n_cpus; i++)
	{
		cpu_history[i].old_state = -1 ;
		cpu_history[i].new_state = -1 ;
	}
}

int wakeup(int cpu)
{
	if (sem.number[cpu] >= n_cpus) return(1) ;

	if (sem.number[cpu] < 0)
	{
	    printf("wakeup: sem.number[cpu] is negative %d\n", sem.number[cpu]);
	    return(1) ;
	}

	switch (sem.state[sem.number[cpu]])
	{
	case P_WAITING:
		set_state(sem.number[cpu], P_TEST2) ;
		break ;
	}

	sem.number[cpu]++ ;
	return(0) ;
}

void sem_fsm(int cpu)
{
	switch (sem.state[cpu])
	{
	case IDLE:
		break ;
	case P_START:
		set_state(cpu, P_GET_LOCK1) ;
		break ;
	case P_GET_LOCK1:
		if (get_lock())
		    set_state(cpu, P_PROCESS) ;
		break ;
	case P_RETRY:
	case P_PROCESS:
		if (decr_cnt() < 0)
		{
		    incr_wtg() ;
		    if (test1())
		    {
			give_lock() ;
			set_state(cpu, P_WAITING) ;
		    }
		    else
			set_state(cpu, P_WAIT_DONE) ;
		}
		else
		{
		    give_lock() ;
		    set_state(cpu, P_ACQUIRED) ;
		}
		break ;
	case P_WAITING:
		/* Hang in this state until awakened by a V operation */
		set_state(cpu, P_WAITING) ;
		break ;
	case P_TEST2:
		if (test2())
		    set_state(cpu, P_WAITING) ;
		else
		    set_state(cpu, P_GET_LOCK2) ;
		break ;
	case P_GET_LOCK2:
		if (get_lock())
		    set_state(cpu, P_WAIT_DONE) ;
		break ;
	case P_WAIT_DONE:
		sem.number[cpu] = norm_exchg() ;
		norm_add(sem.number[cpu]) ;
		set_state(cpu, P_RETRY) ;
		break ;
	case P_ACQUIRED:
		/* Hang in this state until V operation started */
		break ;

	case V_START:
		set_state(cpu, V_INCR_CNT) ;
		break ;
	case V_INCR_CNT:
		if (incr_cnt() <= 0)
			set_state(cpu, V_GET_LOCK) ;
		else
			set_state(cpu, V_COMPLETE) ;
		break ;
	case V_GET_LOCK:
		if (get_lock())
			set_state(cpu, V_PROCESS) ;
		break ;
	case V_PROCESS:
		sem.number[cpu] = norm_exchg() ;
		norm_add(sem.number[cpu]) ;
		sem.number[cpu] = 0 ;		/* cpu number */
		set_state(cpu, V_WAKEUP) ;
		give_lock() ;
		break ;
	case V_WAKEUP:
		set_state(cpu, V_WAKEUP) ;
		if (wakeup(cpu))
			set_state(cpu, V_COMPLETE) ;
		break ;
	case V_COMPLETE:
		break ;
	default:
		printf("sem_fsm: invalid state: %d\n", sem.state[cpu]) ;
		break ;
	}
}

int cpus_done(void)
{
    if (   state_map[V_COMPLETE] == 1
        && state_map[P_WAITING] == 1
	&& state_map[P_ACQUIRED] == 1
       )
    {
	   correct = 1 ;
	   correct_cnt++ ;
	   return(1) ;
    }

    return(0) ;
}

int cpus_hung(void)
{
    if (   state_map[V_COMPLETE] == 1
        && (   state_map[P_WAITING] == n_cpus - 1
	    || state_map[P_ACQUIRED] > 1
           )
       )
    {
	correct = 0 ;
	incorrect_cnt++ ;
	return(1) ;
    }

    return(0) ;
}


void recursive_fsm(int level)
{
    semaphore_t	entry_sem ;
    int		entry_state_map[N_STATES] ;
    int		entry_hist_inx ;
    int		cpu ;
    static int	cases ;

    entry_hist_inx = state_hist_inx ;
    entry_sem = sem ;
    memcpy(entry_state_map, state_map, sizeof(state_map)) ;
    for (cpu = 0; cpu < n_cpus; cpu++, cases++)
    {
	cpu_path[level] = '0' + cpu ;
	cpu_path[level+1] = 0 ;

	sem_fsm(cpu) ;

	if (cpus_done() || cpus_hung() || level == 100)
	{					/* cpus in ending state */
	    if (level == 100)
	    {
		incorrect_cnt++ ;
		cycle_cnt++ ;
	    }
	    else
	    if (   correct == 0
	        || (correct == 1 && n_examples-- > 0)
	       )
	    {
		printf("State History:\n") ;
		dump_state_hist() ;
		printf("\nCPU History:\n") ;
		dump_cpu_hist() ;
	    }
#if 0
	    else
	    {
		printf("%d cases: good %d bad %d cycles %d %s\n",
			cases, correct_cnt, incorrect_cnt, cycle_cnt, cpu_path) ;
		dump_cpu_hist() ;
	    }
#endif
	    fflush(stdout) ;
	}
	else					/* keep going, maybe recurse */
	if (   sem.state[cpu] != entry_sem.state[cpu]
	    || sem.number[cpu] != entry_sem.number[cpu])
	    recursive_fsm(level+1) ;		/* this cpu not idle */

    	/* Now iterate with the next CPU */
	sem = entry_sem ;
	memcpy(state_map, entry_state_map, sizeof(state_map)) ;
	state_hist_inx = entry_hist_inx ;	/* restore hist inx */

	if (cases % 100000 == 0)
	{
	    printf("%d cases: good %d bad %d cycles %d %s\n",
		    cases, correct_cnt, incorrect_cnt, cycle_cnt, cpu_path) ;
	    fflush(stdout) ;
	}
    }

    if (level == 1)			/* finally done */
    {
	printf("%d cases: good %d bad %d cycles %d\n",
		cases, correct_cnt, incorrect_cnt, cycle_cnt) ;
	fflush(stdout) ;
    }
}

void catch_intr(int x)
{
    printf("\nInterrupt\n") ;
    printf("State History:\n") ;
    dump_state_hist() ;
    printf("\nCPU History:\n") ;
    dump_cpu_hist() ;
    exit(0) ;
}

void main(void)
{
#if 1
    signal(SIGINT, catch_intr) ;
    signal(SIGTERM, catch_intr) ;

    init_state_hist() ;
    init_cpu_hist() ;

    state_map[IDLE] = 3 ;		/* 3 cpus in idle state */
    set_state(0, P_START) ;
    set_state(1, P_START) ;
    set_state(2, V_START) ;
    n_cpus = 3 ;			/* number of CPUs we are testing */
    recursive_fsm(1) ;			/* enumerate all cases */

#else
    int		cases ;

    for (cases = 1;;cases++)
    {
	init_state_hist() ;
	init_cpu_hist() ;
	memset(&sem, 0, sizeof(semaphore_t)) ;

	state_map[IDLE] = 3 ;		/* 3 cpus in idle state */
	set_state(0, P_START) ;
	set_state(1, P_START) ;
	set_state(2, V_START) ;
	n_cpus = 3 ;			/* number of CPUs we are testing */
	run_fsms() ;

	if (correct == 0)
	{
	    printf("State History:\n") ;
	    dump_state_hist() ;
	    printf("\nCPU History:\n") ;
	    dump_cpu_hist() ;
	}
	else
	if (cases % 10000 == 0)
	{
	    printf("Tested %d cases\n", cases) ;
	    fflush(stdout) ;
	}
    }
#endif
}
