/*
 * This version of the simulator models semaphores in which the
 * following changes have been made:
 *
 *  1) The code no longer returns to the original calling site for a
 *     retry of the semaphore.
 *  2) The "waiting" field has been changed to a "waking" field.  It is
 *     incremented by the "up()" code when it is about to wake up waiting
 *     processes.  This increment can be done atomically.
 *  3) The semaphore has a "lock" field that is used by the "down()" code
 *     to protect testing and decrementing of the "waking" field.
 */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#define	N_CPUS		8
int	n_cpus ;

/*
 * The simulated semaphore has count and waking field just like
 * the real kernel semaphores.  It also has a state vector each slot
 * of which is used by a different simulated CPU to track the stages
 * of semaphore processing.  An interrupt is viewed as a separate CPU
 * that only does V operations.
 *
 * For simulation purposes, there is only one semaphore and multiple
 * users of it.
 */
typedef struct semaphore
{
	int	count ;			/* counter */
	int	waking ;		/* how many wakeups to be done */
	int	lock ;
	int	state[N_CPUS] ;		/* state of operation */
	int	number[N_CPUS] ;	/* used for waking up cpus */

} semaphore_t ;

/*
 * States of P operation ("down()")
 */
#define	IDLE		'A'		/* no P/V() operation in progress */
#define	P_START		'B'		/* start a P() operation */
#define	P_GET_LOCK1	'C'		/* P: get buzz lock*/
#define	P_PROCESS	'D'		/* P: do most of processing */
#define	P_LOCK		'E'		/* P: get lock and test waking */
#define	P_WAITING	'F'		/* P: give lock and wait on semaphore */
#define	P_TEST2		'G'		/* P: test waking field */
#define	P_DECR_WAKING	'H'		/* P: decr waking field */
#define	P_WAIT_DONE	'I'		/* P: wait complete */
#define	P_ACQUIRED	'J'		/* P: semaphore acquired */
#define	P_RETRY		'K'		/* P: retry the semaphore */
/*
 * States of V operation ("up()")
 */
#define	V_START		'a'		/* start a V() operation */
#define	V_INCR_CNT	'b'		/* V: incr count field */
#define	V_PROCESS	'c'		/* V: incr waking field */
#define	V_START_WAKEUP	'd'		/* V: set up wakeup stte */
#define	V_WAKEUP	'e'		/* V: wake up all waiters */
#define	V_COMPLETE	'f'		/* V: operation complete */

#define	N_STATES	(V_COMPLETE+1)	/* number of states */

semaphore_t	sem ;			/* THE semaphore */


typedef struct
{
	int	cpu ;
	int	old_state ;
	int	new_state ;
	int	count ;
	int	waking ;
	int	number ;

} state_hist_t ;


#define	N_STATE_HIST	1024
state_hist_t	state_history[N_STATE_HIST] ;
int		state_hist_inx ;
state_hist_t	cpu_history[N_CPUS] ;
int		state_map[N_STATES] ;
int		cpu_idle_map[N_CPUS] ;

int		iterations ;
int		correct ;
int		correct_cnt ;
int		incorrect_cnt ;
int		cycle_cnt ;
int		n_examples = 12 ;	/* number of examples to print */

char		cpu_path[2000] = " " ;
char		state_path[2000] = " " ;

char	*state_name(int state)
{
	static char	ascii[32] ;

	switch (state)
	{
	case IDLE:		return("IDLE");

	case P_START:		return("P_START");
	case P_GET_LOCK1:	return("P_GET_LOCK1") ;
	case P_PROCESS:		return("P_PROCESS") ;
	case P_LOCK:		return("P_LOCK") ;
	case P_WAITING:		return("P_WAITING");
	case P_TEST2:		return("P_TEST2");
	case P_DECR_WAKING:	return("P_DECR_WAKING") ;
	case P_WAIT_DONE:	return("P_WAIT_DONE");
	case P_ACQUIRED:	return("P_ACQUIRED");
	case P_RETRY:		return("P_RETRY");

	case V_START:		return("V_START");
	case V_INCR_CNT:	return("V_INCR_CNT");
	case V_PROCESS:		return("V_PROCESS");
	case V_START_WAKEUP:	return("V_START_WAKEUP");
	case V_WAKEUP:		return("V_WAKEUP");
	case V_COMPLETE:	return("V_COMPLETE");

	default:
		sprintf(ascii, "State=%d", state) ;
		return(ascii) ;
	}
}

void print_states(void)
{
    int		 i ;
    char	*p ;

    for (i = 0; i < N_STATES; i++)
    {
	p = state_name(i) ;
	if (strncmp(p, "State=", 6) != 0)
	    printf("%c = %s\n", i, p) ;
    }

    printf("\n") ;
}

int get_lock(void)
{
    if (sem.lock == 0)
    {
	sem.lock = 1 ;
	return(1) ;			/* got lock */
    }

    return(0) ;				/* did not get lock */
}

void give_lock(void)
{
    sem.lock = 0 ;
}

int decr_cnt(void)
{
	return(--sem.count) ;
}


int test2(void)
{
	if (sem.waking <= 0)
	    return(1) ;				/* go back to sleep */

	return(0) ;
}

int incr_cnt(void)
{
	return(++sem.count) ;
}


void incr_waking(void)
{
	sem.waking++ ;
}

void decr_waking(void)
{
	--sem.waking ;
}

void save_state_hist(state_hist_t *p, int cpu,
		     int old_state, int new_state, int *inx)
{
	p->cpu       = cpu ;
	p->old_state = old_state ;
	p->new_state = new_state ;
	p->count     = sem.count ;
	p->waking    = sem.waking ;
	p->number    = sem.number[cpu] ;

	if (inx != NULL)
		*inx = (*inx + 1) % N_STATE_HIST ;

}

void set_state(int cpu, int state)
{
	save_state_hist(&state_history[state_hist_inx], 
			cpu, sem.state[cpu], state, &state_hist_inx) ;

	save_state_hist(&cpu_history[cpu], 
			cpu, sem.state[cpu], state, NULL) ;

	state_map[sem.state[cpu]]-- ;		/* one fewer in that state */
	state_map[state]++ ;			/* one more in this state */
	sem.state[cpu] = state ;
}

void print_state_hist(state_hist_t *p, int index)
{
	if (p->old_state == -1) return ;

	printf("cpu %d: index=%04d count=%d waking=%d number=%d %s(%c)",
		p->cpu,
		++index,
		p->count,
		p->waking,
		p->number,
		state_name(p->old_state), p->old_state) ;

	if (p->new_state >= 0)
		printf(" -> %s(%c)", state_name(p->new_state), p->new_state) ;

	printf("\n") ;
} 

void dump_state_hist(void)
{
	int	i ;

	for (i = 0;
	     i < state_hist_inx;
	     i++
	    )
        {
		print_state_hist(&state_history[i], i) ;
        }

	fflush(stdout) ;
}

void dump_cpu_hist(void)
{
	int	i ;

	for (i = 0; i < n_cpus; i++)
        {
		print_state_hist(&cpu_history[i], i) ;
        }

	fflush(stdout) ;
}

void init_state_hist(void)
{
	int	i ;

	for (i = 0; i < N_STATE_HIST; i++)
	{
		state_history[i].old_state = -1 ;
		state_history[i].new_state = -1 ;
	}

	state_hist_inx = 0 ;
	memset(state_map, 0, sizeof(state_map)) ;
}

void init_cpu_hist(void)
{
	int	i ;

	for (i = 0; i < n_cpus; i++)
	{
		cpu_history[i].old_state = -1 ;
		cpu_history[i].new_state = -1 ;
	}
}

int wakeup(int cpu)
{
	if (sem.number[cpu] >= n_cpus) return(1) ;

	if (sem.number[cpu] < 0)
	{
	    printf("wakeup: sem.number[cpu] is negative %d\n", sem.number[cpu]);
	    return(1) ;
	}

	switch (sem.state[sem.number[cpu]])
	{
	case P_WAITING:
		set_state(sem.number[cpu], P_LOCK) ;
		break ;
	}

	sem.number[cpu]++ ;
	if (sem.number[cpu] >= n_cpus) return(1) ;
	return(0) ;
}

void sem_fsm(int cpu)
{
	switch (sem.state[cpu])
	{
	case IDLE:
		break ;
	case P_START:
		set_state(cpu, P_PROCESS) ;
		break ;
	case P_GET_LOCK1:
		if (get_lock())
		    set_state(cpu, P_PROCESS) ;
		break ;
	case P_RETRY:
	case P_PROCESS:
		if (decr_cnt() < 0)
		{
		    set_state(cpu, P_LOCK) ;
		}
		else
		{
		    set_state(cpu, P_ACQUIRED) ;
		}
		break ;
	case P_WAITING:
		/* Hang in this state until awakened by a V operation */
		set_state(cpu, P_WAITING) ;
		break ;
	case P_LOCK:
		if (get_lock())
		    set_state(cpu, P_TEST2) ;
		break ;
	case P_TEST2:
		if (test2())			/* true => keep sleeping */
		{
		    give_lock() ;
		    set_state(cpu, P_WAITING) ;
		}
		else
		    set_state(cpu, P_DECR_WAKING) ;
		break ;
	case P_DECR_WAKING:
		decr_waking() ;
		give_lock() ;
		set_state(cpu, P_ACQUIRED) ;
		break ;
	case P_WAIT_DONE:
		break ;
	case P_ACQUIRED:
		/* Hang in this state until V operation started */
		break ;

	case V_START:
		set_state(cpu, V_INCR_CNT) ;
		break ;
	case V_INCR_CNT:
		if (incr_cnt() <= 0)
			set_state(cpu, V_PROCESS) ;
		else
			set_state(cpu, V_COMPLETE) ;
		break ;
	case V_PROCESS:
		incr_waking() ;
		set_state(cpu, V_START_WAKEUP) ;
		break ;
	case V_START_WAKEUP:
		sem.number[cpu] = 0 ;		/* cpu number */
		set_state(cpu, V_WAKEUP) ;
		break ;
	case V_WAKEUP:
		set_state(cpu, V_WAKEUP) ;
		if (wakeup(cpu))
			set_state(cpu, V_COMPLETE) ;
		break ;
	case V_COMPLETE:
		break ;
	default:
		printf("sem_fsm: invalid state: %d\n", sem.state[cpu]) ;
		break ;
	}
}

int cpus_done(void)
{
    if (   state_map[V_COMPLETE] == 1
        && state_map[P_WAITING] == 1
	&& state_map[P_ACQUIRED] == 1
       )
    {
	if (sem.count == -1 && sem.waking == 0)
	{
	   correct = 1 ;
	   correct_cnt++ ;
	   return(1) ;
	}

	printf("\nCPUs are done but semaphore is in improper ending state\n");
	correct = 0 ;
	incorrect_cnt++ ;
	return(1) ;
    }

    return(0) ;
}

int cpus_hung(void)
{
    if (   state_map[V_COMPLETE] == 1
        && (   state_map[P_WAITING] == n_cpus - 1
	    || state_map[P_ACQUIRED] > 1
           )
       )
    {
	correct = 0 ;
	incorrect_cnt++ ;
	return(1) ;
    }

    return(0) ;
}


void recursive_fsm(int level, int *idle_cpus)
{
    semaphore_t	entry_sem ;
    int		entry_state_map[N_STATES] ;
    int		entry_hist_inx ;
    int		cpu ;
    int		cpu_idle[N_CPUS] ;
    static int	cases ;

    entry_hist_inx = state_hist_inx ;
    entry_sem = sem ;
    memcpy(entry_state_map, state_map, sizeof(state_map)) ;
    memcpy(cpu_idle, idle_cpus, sizeof(cpu_idle)) ;
    for (cpu = 0; cpu < n_cpus; cpu++, cases++)
    {
	cpu_path[level] = '0' + cpu ;
	cpu_path[level+1] = 0 ;

	sem_fsm(cpu) ;

	state_path[level] = sem.state[cpu] ;	/* ending state of FSM */
	state_path[level+1] = 0 ;
	if (cpus_done() || cpus_hung() || level == 100)
	{					/* cpus in ending state */
	    if (level == 100)
	    {
		incorrect_cnt++ ;
		cycle_cnt++ ;
	    }
	    else
	    if (   correct == 0
	        || (   correct == 1
	            && (   correct_cnt == 1
	                || correct_cnt == 50000
	                || correct_cnt == 100000
	                || correct_cnt == 150000
	                || correct_cnt == 200000
	                || correct_cnt == 250000
	                || correct_cnt == 300000
	                || correct_cnt == 383200
		       )
	           )
	       )
	    {
		if (correct == 1)
		    printf("\nCorrect ending condition:\n") ;
		else
		    printf("\nIncorrect ending condition:\n") ;

		printf("State Path: %s\n", state_path) ;
		printf("  CPU Path: %s\n", cpu_path) ;
		printf("State History:\n") ;
		dump_state_hist() ;
		printf("\nCPU History:\n") ;
		dump_cpu_hist() ;
	    }

	    fflush(stdout) ;
	}
	else					/* keep going, maybe recurse */
	{
	    if (   sem.state[cpu] == entry_sem.state[cpu]
		&& sem.number[cpu] == entry_sem.number[cpu])
	    {					/* idle CPU */
		cpu_idle[cpu] = sem.state[cpu] ;	/* save idling state */
	    }
	    else
	    {
		if (cpu_idle[cpu] != sem.state[cpu])	/* CPU changed state */
		{
		    cpu_idle[cpu] = 0 ;			/* not idle */
		    recursive_fsm(level+1, cpu_idle) ;	/* recurse */
		}
	    }
	}

    	/* Now iterate with the next CPU */
	sem = entry_sem ;
	memcpy(state_map, entry_state_map, sizeof(state_map)) ;
	state_hist_inx = entry_hist_inx ;	/* restore hist inx */

	if (cases % 100000 == 0)
	{
	    printf("%d cases: good %d bad %d cycles %d\n",
		    cases, correct_cnt, incorrect_cnt, cycle_cnt) ;
	    printf("\tState Path: %s\n", state_path) ;
	    printf("\t  CPU Path: %s\n", cpu_path) ;
	    fflush(stdout) ;
	}
    }

    if (level == 1)			/* finally done */
    {
	printf("%d cases: good %d bad %d cycles %d\n",
		cases, correct_cnt, incorrect_cnt, cycle_cnt) ;
	fflush(stdout) ;
    }
#if 0
    else
    {
	printf("\nExit recursive_fsm:\n") ;
	printf("%d cases: good %d bad %d cycles %d\n",
		cases, correct_cnt, incorrect_cnt, cycle_cnt) ;
	printf("\tState Path: %s\n", state_path) ;
	printf("\t  CPU Path: %s\n", cpu_path) ;
	printf("\nCPU History:\n") ;
	dump_cpu_hist() ;
	fflush(stdout) ;
    }
#endif
}


void catch_intr(int x)
{
    printf("\nInterrupt\n") ;
    printf("State History:\n") ;
    dump_state_hist() ;
    printf("\nCPU History:\n") ;
    dump_cpu_hist() ;
    exit(0) ;
}

void main(void)
{
    signal(SIGINT, catch_intr) ;
    signal(SIGTERM, catch_intr) ;

    print_states() ;
    init_state_hist() ;
    init_cpu_hist() ;

    state_map[IDLE] = 3 ;		/* 3 cpus in idle state */
    set_state(0, P_START) ;
    set_state(1, P_START) ;
    set_state(2, V_START) ;
    n_cpus = 3 ;			/* number of CPUs we are testing */
    recursive_fsm(1, cpu_idle_map) ;	/* enumerate all cases */

}
